const http = require('http');
const { Semantics, Semantic } = require('./src/semantics');
const { Generators, Generator } = require('./src/generators');
const { Digraph } = require('./src/digraph');
const fetch = require('node-fetch')
const _ = require('underscore');
const base64 = require('base-64')
const fs = require('fs')
const { ArgumentParser } = require('argparse')

const processContexts = (contexts, params) => {
  contextsPrime = []
  generateds = []
  for (context of contexts) {
    const result = processContext(context, params)
    contextsPrime.push(result.context)
    generateds.push(result.generated)
  }
  return { contexts: contextsPrime, generated: generateds }
}

const processContext = (context, { objects = {}, generators = [], semantics = [], config }) => {
  // map to hash
  config = config || {}
  if (config.config) {
    config = config;
  }
  const response = {} // NA but passed in
  generators = new Generators(generators.map( (g) => new Generator(g[0], g[1], g[2]) ) )
  semantics = new Semantics(semantics.map( (g) => new Semantic(g[0], g[1], g[2]) ) )

  context = semantics.apply(objects, context, config, response)
  const generated = generators.apply(objects, context, config, response)[0]

  return { context, generated }
}

const writeTest = (fn, query, generated, result) => {
  let tests = {}
  if (fs.existsSync(fn)) {
    tests = JSON.parse(fs.readFileSync(fn))
  }
  tests[query] = {generated, result}
  fs.writeFileSync(fn, JSON.stringify(tests, null, 2),  {"encoding": "utf8", "flag": 'w+'})
}

const _process = (url, key, config, { writeTests, testsFN } = {}) => {
  retries = 2
  writeTests = writeTests || false

  config.initialize()
  let data = Object.assign( { key, version: '3' }, config.config );
  const generators = new Generators((data.generators || []).map( (g) => new Generator(g[0], g[1], g[2]) ) )
  delete data['generators']
  const semantics = new Semantics((data.semantics || []).map( (g) => new Semantic(g[0], g[1], g[2]) ) )
  delete data['semantics']

  if (config.config.utterances.length == 0 || config.config.utterances.some( (u) => u.length == 0 )) {
    throw "No utterances were specified"
  }

  const doWithRetries = (n) => {
    return fetch(`${url}/process`, {
      method: "POST",
      body: JSON.stringify(data),
      timeout: 1000 * 60 * 5, // it does not respect this timeout so that's why I have the retries
      headers: {
        mode: "no-cors",
        "Content-Type": "application/json"
      },
      }).then( result => {
        if (result.ok) {
          return result.json()
        }
        if (result.status == 504){
          if (n == 0) {
            throw `Error ${result.status} - ${result.statusText}`;
          } else {
            return doWithRetries(n-1)
          }
        }
        if (result.status == 503) {
          throw `Error ${result.status} - ${result.statusText}. If you are using the developer version try restarting your entodicton server. If takes a few minutes to be up.`;
        } else {
          throw `Error ${result.status} - ${result.statusText}`;
        }
      })
    }

  let todo = doWithRetries(retries);
  return todo
    .then( json => {
      if (json.status != 200) {
        throw json;
      }
      if (json.results) {
        let resultsPrime = []
        let generatedsPrime = []
        for (let i = 0; i < json.results.length; ++i ) {
          let resultPrime = []
          let generatedPrime = []
          const result = json.results[i];
          for (let j = 0; j < result.length; ++j) {
            const context = result[j];
            let contextPrime = context
            if (!config.get('skipSemantics')) {
              contextPrime = semantics.apply(config.get("objects"), context, config, json)
            }
            try {
              const generated = generators.apply(config.get("objects"), contextPrime, config, json)[0]
              resultPrime.push(contextPrime)
              generatedPrime.push(generated)
            } catch( e ) {
              e.context = contextPrime
              e.logs = json.logs
              if (json.trace) {
                e.trace = json.trace;
              }
              throw e
            }
          }
          resultsPrime.push(resultPrime);
          generatedsPrime.push(generatedPrime);
          json.results = resultsPrime;
          json.generated = generatedsPrime;

          if (writeTests) {
            for (let i = 0; i < json.generated.length; ++i) {
              writeTest(testsFN, config.get("utterances")[i], json.generated[i], json.results[i])
            }
          }
        }
      }
      if (json.contexts) {
        json.results = json.results[0]
        json.generated = json.generated[0]
      }
      return json;
    });
}

const runTest = (url, key, config, test, expected) => {
  config.config.utterances = [test]
  return _process(url, key, config).then( (result) => {
    if (JSON.stringify(result.generated) != JSON.stringify([expected.generated]) || JSON.stringify(result.results) != JSON.stringify([expected.result])) {
      return { 
        utterance: test, 
        expected: { generated: expected.generated, results: expected.result }, 
        actual: { generated: result.generated[0], results: result.results[0] } 
      }
    }
  })
}

const runTestsHelper = (url, key, config, tests, todo, failed, stopAtFirstError) => {
  if (todo.length == 0) {
    return failed;
  }
  const test = todo.pop()
  return runTest(url, key, config, test, tests[test]).then( (result) => {
    if (result != null) {
      failed.push(result)
      if (stopAtFirstError) {
        return failed
      }
    }
    return runTestsHelper(url, key, config, tests, todo, failed, stopAtFirstError)
  })
}

const runTests = (url, key, config, testFile, { stopAtFirstError = false } = {}) => {
  const tests = JSON.parse(fs.readFileSync(testFile))
  return runTestsHelper(url, key, config, tests, Object.keys(tests), [], stopAtFirstError)
};

const showExamples = (testFile) => {
  const tests = JSON.parse(fs.readFileSync(testFile))
  Object.keys(tests).forEach( (test) => console.log(test) )
}

const showInfo = (name, description, config, testFile) => {
  const tests = JSON.parse(fs.readFileSync(testFile))
  const includes = config.configs.slice(1).map( (km) => km.config.name )
  const examples = Object.keys(tests)
  console.log(JSON.stringify({ name, description, examples, includes }, null, 2))
}

const matching = (actual, expected) => {
  if (!_.isEqual(actual, expected)) {
    return false;
  }
  return true;
}

const submitBugToAPI = (subscription_id, subscription_password, config) => {
  console.log("********* Submitting bug *********")
  const body = { description: config.config.description, config: config.config };
  //fetch('http://localhost:5000/bug', {
  fetch('https://thinktelligence.com/api/bug', {
      method: "POST",
      body: JSON.stringify(body),
      headers: {
        mode: "no-cors", // Type of mode of the request
        "Content-Type": "application/json", // request content type
        "Authorization": 'Basic ' + base64.encode(subscription_id + ":" + subscription_password)
      } 
    }).then( (result) => result.json() )
      .then( (json) => {
      if (json.error) {
        console.log(`Error submitting the bug: ${json.error}`);
      } else {
        console.log(`New bug id id ${json.id}`);
      }
    });
}

const submitBug = (subscription_id, subscription_password, entodicton_url, entodicton_key, config, retries = 2) => {
  const properties = ['expected_results', 'expected_generated', 'utterances']
  properties.forEach( (property) => {
    if (!config.get('expected_results') ) {
      throw "Missing property expected_results";
    }
  });

  return _process(url, key, config)
    .then( (responses) => {
      let hasError = false;
      if (!matching(responses.results, config.config.expected_results)) {
        console.log('JSON does not match');
        console.log('actual', JSON.stringify(responses.results))
        console.log('expected', JSON.stringify(config.config.expected_results))
        hasError = true;
      }
      if (!matching(responses.generated, config.config.expected_generated)) {
        console.log('Generated does not match');
        console.log('actual', JSON.stringify(responses.generated))
        console.log('expected', JSON.stringify(config.config.expected_generated))
        hasError = true;
      }

      if (hasError) {
        submitBugToAPI(subscription_id, subscription_password, config)
      } else {
        throw "\n\n *** The expected results matched the generated results so the bug was not submitted since its working. ***\n\n:"
      }
    })
    .catch( (error) => { 
      console.log('Error', error)
      throw error;
    })
}

const isEntryPoint = () => {
  return require.main === module;
}

// setup is run before test or debug. 
const knowledgeModule = ({url, name, description, key, config, setup=()=>{}, test, process : processResults, module, stopAtFirstFailure = true, isProcess } = {}) => {
  config.name = name
  config.description = description
  if (isProcess) {
    config.tests = JSON.parse(fs.readFileSync(test))
    setup();
    const parser = new ArgumentParser({
        description: 'Entodicton knowledge module'
    });
     
    parser.add_argument('-t', '--test', { action: 'store_true', help: 'Run the tests. Create tests by running with the --query or --loop + --save flag' });
    parser.add_argument('-l', '--loop', { action: 'store_true', help: 'Run a loop so that multiply queries may be run' });
    parser.add_argument('-i', '--info', { action: 'store_true', help: 'Print meta-data for the module' });
    parser.add_argument('-q', '--query', { help: 'Run the specified query' });
    parser.add_argument('-s', '--save', { action: 'store_true', help: 'When running with the --query flag this will save the current run to the test file' });
      
    args = parser.parse_args();

    if (args.info) {
      showInfo(name, description, config, test)
    } else if (args.test) {
      if (typeof test == 'string') {
        runTests(url, key, config, test, { stopAtFirstError: true }).then( (failures) => {
          console.log("failures", JSON.stringify(failures, null, 2))
          console.log("failed queries")
          failures.forEach( (f) => console.log("  " + f.utterance) )
        })
      } else {
        test()
      }
    } else if (args.loop) {
      const readline = require('readline').createInterface({ input: process.stdin, output: process.stdout });
      const f = () => readline.question('Enter query? (newline to quit) ', query => {
        query = query.trim()
        if (query.length == 0) {
          return readline.close()
        }
        config.config.utterances = [query];
        const promise = processResults(_process(url, key, config, { writeTests: args.save, testsFN: test }))
        if (!('then' in promise)) {
          throw "Return a promise from process in the definition of knowledgeModule"
        }
        promise.then( () => f() )
      });
      f();
    } else if (args.query) {
      config.config.utterances = [args.query];
      processResults(_process(url, key, config, { writeTests: args.save, testsFN: test }))
    }
  } else {
    module()
  }
}

module.exports = { 
  process: _process, 
  submitBug, 
  processContext, 
  processContexts, 
  runTests, 
  knowledgeModule,
  Semantics,
  Semantic,
  Generators,
  Generator,
  Digraph
}
